# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest as _ut


class ensemble_test_case(_ut.TestCase):
    def test_batch(self):
        from . import (
            ensemble_propagate_until_batch,
            ensemble_propagate_for_batch,
            ensemble_propagate_grid_batch,
            make_vars,
            sin,
            taylor_adaptive_batch,
        )
        from .callback import angle_reducer
        import numpy as np

        x, v = make_vars("x", "v")

        # Use a pendulum for testing purposes.
        sys = [(x, v), (v, -9.8 * sin(x))]

        algos = ["thread", "process"]

        ta = taylor_adaptive_batch(sys=sys, state=[[0.0] * 4] * 2)

        ics = np.zeros((10, 2, 4))
        for i in range(10):
            ics[i, 0] = [
                0.05 + i / 100,
                0.051 + i / 100,
                0.052 + i / 100,
                0.053 + i / 100.0,
            ]
            ics[i, 0] = [
                0.025 + i / 100,
                0.026 + i / 100,
                0.027 + i / 100,
                0.028 + i / 100.0,
            ]

        # propagate_until().
        def gen(ta, idx):
            ta.set_time(0.0)
            ta.state[:] = ics[idx]

            return ta

        for algo in algos:
            if algo == "thread":
                ret = ensemble_propagate_until_batch(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8
                )
            elif algo == "process":
                ret = ensemble_propagate_until_batch(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8, chunksize=3
                )

            self.assertEqual(len(ret), 10)

            for i in range(10):
                ta.set_time(0.0)
                ta.state[:] = ics[i]
                loc_ret = ta.propagate_until(20.0)

                for j in range(4):
                    self.assertAlmostEqual(ret[i][0].time[j], 20.0)
                self.assertTrue(np.all(ta.state == ret[i][0].state))
                self.assertTrue(ret[i][1] is None)
                self.assertTrue(np.all(ta.time == ret[i][0].time))
                self.assertEqual(ta.propagate_res, ret[i][0].propagate_res)

            # Run a test with c_output too.
            if algo == "thread":
                ret = ensemble_propagate_until_batch(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8, c_output=True
                )
            elif algo == "process":
                ret = ensemble_propagate_until_batch(
                    ta,
                    20.0,
                    10,
                    gen,
                    algorithm=algo,
                    max_workers=8,
                    chunksize=3,
                    c_output=True,
                )

            self.assertEqual(len(ret), 10)

            for i in range(10):
                ta.set_time(0.0)
                ta.state[:] = ics[i]
                loc_ret = ta.propagate_until(20.0, c_output=True)

                for j in range(4):
                    self.assertAlmostEqual(ret[i][0].time[j], 20.0)
                self.assertTrue(np.all(ta.state == ret[i][0].state))
                self.assertFalse(ret[i][1] is None)
                self.assertTrue(np.all(ta.time == ret[i][0].time))
                self.assertEqual(ta.propagate_res, ret[i][0].propagate_res)

                self.assertTrue(np.all(loc_ret[0](5.0) == ret[i][1](5.0)))

        # propagate_for().
        def gen(ta, idx):
            ta.set_time(10.0)
            ta.state[:] = ics[idx]

            return ta

        for algo in algos:
            if algo == "thread":
                ret = ensemble_propagate_for_batch(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8
                )
            elif algo == "process":
                ret = ensemble_propagate_for_batch(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8, chunksize=3
                )

            self.assertEqual(len(ret), 10)

            for i in range(10):
                ta.set_time(10.0)
                ta.state[:] = ics[i]
                loc_ret = ta.propagate_for(20.0)

                for j in range(4):
                    self.assertAlmostEqual(ret[i][0].time[j], 30.0)
                self.assertTrue(np.all(ta.state == ret[i][0].state))
                self.assertTrue(ret[i][1] is None)
                self.assertTrue(np.all(ta.time == ret[i][0].time))
                self.assertEqual(ta.propagate_res, ret[i][0].propagate_res)

        # propagate_grid().
        grid = np.linspace(0.0, 20.0, 80)

        splat_grid = np.repeat(grid, 4).reshape(-1, 4)

        def gen(ta, idx):
            ta.set_time(0.0)
            ta.state[:] = ics[idx]

            return ta

        for algo in algos:
            if algo == "thread":
                ret = ensemble_propagate_grid_batch(
                    ta, grid, 10, gen, algorithm=algo, max_workers=8
                )
            elif algo == "process":
                ret = ensemble_propagate_grid_batch(
                    ta, grid, 10, gen, algorithm=algo, max_workers=8, chunksize=3
                )

            self.assertEqual(len(ret), 10)

            for i in range(10):
                ta.set_time(0.0)
                ta.state[:] = ics[i]
                loc_ret = ta.propagate_grid(splat_grid)

                self.assertTrue(np.all(loc_ret[1] == ret[i][2]))

                for j in range(4):
                    self.assertAlmostEqual(ret[i][0].time[j], 20.0)
                self.assertTrue(np.all(ta.state == ret[i][0].state))
                self.assertTrue(np.all(ta.time == ret[i][0].time))
                self.assertEqual(ta.propagate_res, ret[i][0].propagate_res)

        # Check that callbacks are deep-copied in thread-based
        # ensemble propagations.

        class step_cb:
            def __call__(_, ta):
                self.assertNotEqual(id(_), _.orig_id)

                return True

        cb = step_cb()
        cb.orig_id = id(cb)

        def gen(ta, idx):
            ta.set_time(0.0)
            ta.state[:] = ics[idx]

            return ta

        ret = ensemble_propagate_for_batch(
            ta, 20.0, 10, gen, algorithm="thread", max_workers=8, callback=cb
        )

        for r in ret:
            self.assertTrue(isinstance(r[2], step_cb))

        # Test the list overload too.
        ret = ensemble_propagate_for_batch(
            ta,
            20.0,
            10,
            gen,
            algorithm="thread",
            max_workers=8,
            callback=[cb, angle_reducer([x])],
        )

        for r in ret:
            self.assertTrue(isinstance(r[2], list))
            self.assertTrue(isinstance(r[2][0], step_cb))
            self.assertTrue(isinstance(r[2][1], angle_reducer))
            self.assertEqual(len(r[2]), 2)

        # Test s11n machinery in multi-processing situations.
        ret = ensemble_propagate_for_batch(
            ta,
            20.0,
            10,
            gen,
            algorithm="process",
            callback=[cb, angle_reducer([x])],
        )

        for r in ret:
            self.assertTrue(isinstance(r[2], list))
            self.assertTrue(isinstance(r[2][0], step_cb))
            self.assertTrue(isinstance(r[2][1], angle_reducer))
            self.assertEqual(len(r[2]), 2)

    def test_scalar(self):
        from . import (
            ensemble_propagate_until,
            ensemble_propagate_for,
            ensemble_propagate_grid,
            make_vars,
            sin,
            taylor_adaptive,
            taylor_outcome,
        )
        from .callback import angle_reducer
        import numpy as np

        x, v = make_vars("x", "v")

        # Use a pendulum for testing purposes.
        sys = [(x, v), (v, -9.8 * sin(x))]

        algos = ["thread", "process"]

        ta = taylor_adaptive(sys=sys, state=[0.0] * 2)

        ics = np.array([[0.05, 0.025]] * 10)
        for i in range(10):
            ics[i] += i / 100.0

        # propagate_until().
        def gen(ta, idx):
            ta.time = 0.0
            ta.state[:] = ics[idx]

            return ta

        for algo in algos:
            if algo == "thread":
                ret = ensemble_propagate_until(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8
                )
            elif algo == "process":
                ret = ensemble_propagate_until(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8, chunksize=3
                )

            self.assertEqual(len(ret), 10)

            self.assertTrue(all([_[1] == taylor_outcome.time_limit for _ in ret]))

            for i in range(10):
                ta.time = 0.0
                ta.state[:] = ics[i]
                loc_ret = ta.propagate_until(20.0)

                self.assertAlmostEqual(ret[i][0].time, 20.0)
                self.assertTrue(np.all(ta.state == ret[i][0].state))
                self.assertEqual(loc_ret, ret[i][1:])
                self.assertEqual(ta.time, ret[i][0].time)

            # Run a test with c_output too.
            if algo == "thread":
                ret = ensemble_propagate_until(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8, c_output=True
                )
            elif algo == "process":
                ret = ensemble_propagate_until(
                    ta,
                    20.0,
                    10,
                    gen,
                    algorithm=algo,
                    max_workers=8,
                    chunksize=3,
                    c_output=True,
                )

            self.assertEqual(len(ret), 10)

            self.assertTrue(all([_[1] == taylor_outcome.time_limit for _ in ret]))

            for i in range(10):
                ta.time = 0.0
                ta.state[:] = ics[i]
                loc_ret = ta.propagate_until(20.0, c_output=True)

                self.assertAlmostEqual(ret[i][0].time, 20.0)
                self.assertTrue(np.all(ta.state == ret[i][0].state))
                self.assertEqual(loc_ret[:-2], ret[i][1:-2])
                self.assertEqual(ta.time, ret[i][0].time)

                self.assertTrue(np.all(loc_ret[-2](5.0) == ret[i][-2](5.0)))

        # propagate_for().
        def gen(ta, idx):
            ta.time = 10.0
            ta.state[:] = ics[idx]

            return ta

        for algo in algos:
            if algo == "thread":
                ret = ensemble_propagate_for(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8
                )
            elif algo == "process":
                ret = ensemble_propagate_for(
                    ta, 20.0, 10, gen, algorithm=algo, max_workers=8, chunksize=3
                )

            self.assertEqual(len(ret), 10)

            self.assertTrue(all([_[1] == taylor_outcome.time_limit for _ in ret]))

            for i in range(10):
                ta.time = 10.0
                ta.state[:] = ics[i]
                loc_ret = ta.propagate_for(20.0)

                self.assertAlmostEqual(ret[i][0].time, 30.0)
                self.assertTrue(np.all(ta.state == ret[i][0].state))
                self.assertEqual(loc_ret, ret[i][1:])
                self.assertEqual(ta.time, ret[i][0].time)

        # propagate_grid().
        grid = np.linspace(0.0, 20.0, 80)

        def gen(ta, idx):
            ta.time = 0.0
            ta.state[:] = ics[idx]

            return ta

        for algo in algos:
            if algo == "thread":
                ret = ensemble_propagate_grid(
                    ta, grid, 10, gen, algorithm=algo, max_workers=8
                )
            elif algo == "process":
                ret = ensemble_propagate_grid(
                    ta, grid, 10, gen, algorithm=algo, max_workers=8, chunksize=3
                )

            self.assertEqual(len(ret), 10)

            self.assertTrue(all([_[1] == taylor_outcome.time_limit for _ in ret]))

            for i in range(10):
                ta.time = 0.0
                ta.state[:] = ics[i]
                loc_ret = ta.propagate_grid(grid)

                self.assertAlmostEqual(ret[i][0].time, 20.0)
                self.assertTrue(np.all(ta.state == ret[i][0].state))
                self.assertEqual(loc_ret[:-1], ret[i][1:-1])
                self.assertTrue(np.all(loc_ret[-1] == ret[i][-1]))
                self.assertEqual(ta.time, ret[i][0].time)

        # Error handling.
        with self.assertRaises(TypeError) as cm:
            ensemble_propagate_until(ta, 20.0, "a", gen)
        self.assertTrue(
            "The n_iter parameter must be an integer, but an object of type"
            in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            ensemble_propagate_until(ta, 20.0, -1, gen)
        self.assertTrue(
            "The n_iter parameter must be non-negative" in str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            ensemble_propagate_until(ta, [20.0], 10, gen)
        self.assertTrue(
            "Cannot perform an ensemble propagate_until/for(): the final epoch/time"
            " interval must be a scalar, not an iterable object" in str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            ensemble_propagate_for(ta, [20.0], 10, gen)
        self.assertTrue(
            "Cannot perform an ensemble propagate_until/for(): the final epoch/time"
            " interval must be a scalar, not an iterable object" in str(cm.exception)
        )

        with self.assertRaises(ValueError) as cm:
            ensemble_propagate_grid(ta, [[20.0, 20.0]], 10, gen)
        self.assertTrue(
            "Cannot perform an ensemble propagate_grid(): the input time grid must be"
            " one-dimensional, but instead it has 2 dimensions" in str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            ensemble_propagate_until(ta, 20.0, 10, gen, max_delta_t=[10])
        self.assertTrue(
            'Cannot perform an ensemble propagate_until/for/grid(): the "max_delta_t"'
            " argument must be a scalar, not an iterable object" in str(cm.exception)
        )

        # NOTE: check that the chunksize option is not recognised
        # in threaded mode.
        with self.assertRaises(TypeError) as cm:
            ensemble_propagate_until(ta, 20.0, 10, gen, chunksize=1)

        # Check that callbacks are deep-copied in thread-based
        # ensemble propagations.

        class step_cb:
            def __call__(_, ta):
                self.assertNotEqual(id(_), _.orig_id)

                return True

        cb = step_cb()
        cb.orig_id = id(cb)

        def gen(ta, idx):
            ta.time = 0.0
            ta.state[:] = ics[idx]

            return ta

        ret = ensemble_propagate_for(
            ta, 20.0, 10, gen, algorithm="thread", max_workers=8, callback=cb
        )

        for r in ret:
            self.assertTrue(isinstance(r[-1], step_cb))

        # Test the list overload too.
        ret = ensemble_propagate_for(
            ta,
            20.0,
            10,
            gen,
            algorithm="thread",
            max_workers=8,
            callback=[cb, angle_reducer([x])],
        )

        for r in ret:
            self.assertTrue(isinstance(r[-1], list))
            self.assertTrue(isinstance(r[-1][0], step_cb))
            self.assertTrue(isinstance(r[-1][1], angle_reducer))
            self.assertEqual(len(r[-1]), 2)

        # Test s11n machinery in multi-processing situations.
        ret = ensemble_propagate_for(
            ta,
            20.0,
            10,
            gen,
            algorithm="process",
            callback=[cb, angle_reducer([x])],
        )

        for r in ret:
            self.assertTrue(isinstance(r[-1], list))
            self.assertTrue(isinstance(r[-1][0], step_cb))
            self.assertTrue(isinstance(r[-1][1], angle_reducer))
            self.assertEqual(len(r[-1]), 2)



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .. import core as _core

_lst = list(filter(lambda name: name.startswith("_callback_"), dir(_core)))

for _name in _lst:
    exec(f"from ..core import {_name} as {_name[10:]}")

del _core
del _lst
del _name



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

from .. import core as _core, expression
from typing import Union
import numpy

_lst = list(filter(lambda name: name.startswith("_model_"), dir(_core)))

for _name in _lst:
    exec(f"from ..core import {_name} as {_name[7:]}")
    # NOTE: we also try to change the __name__ of the current object.
    # This will work only for exposed classes, not functions, hence
    # the try/except block. We do this because otherwise sphinx
    # will document classes as aliases for the implementation detail
    # objects with the weird name.
    try:
        exec(f"{_name[7:]}.__name__ = '{_name[7:]}'")
    except AttributeError:
        pass

# NOTE:redefine the models which are exposed as attributes,
# so that we can document them.

delta_tt_tai: expression = delta_tt_tai
"""
Difference between TT and TAI.

.. versionadded:: 7.3.0

This expression is a constant representing the difference between `terrestrial time (TT) <https://en.wikipedia.org/wiki/Terrestrial_Time>`__
and `international atomic time (TAI) <https://en.wikipedia.org/wiki/International_Atomic_Time>`__.

This difference amounts to exactly 32.184 SI seconds.

"""

del expression


def sgp4_propagator(
    sat_list: Union[list, numpy.ndarray],
    diff_order: int = 0,
    **kwargs,
):
    """
    Construct an SGP4 propagator.

    .. versionadded:: 5.1.0

    .. versionadded:: 7.0.0

       This function now also accepts *sat_list* as a NumPy array.

    .. note::

       A :ref:`tutorial <tut_sgp4_propagator>` explaining the use of this function
       is available.

    This function will construct an SGP4 propagator from the input arguments.

    The only mandatory argument is *sat_list*, which must be either a list
    of general perturbations element sets (GPEs) represented as ``Satrec`` objects from the
    `sgp4 Python module <https://pypi.org/project/sgp4/>`__, or a 2D array.

    In the former case, the GPE data is taken directly from the ``Satrec`` objects.
    In the latter case, *sat_list* is expected to be a 9 x ``n`` C-style contiguous
    array, where ``n`` is the total number of satellites and the rows contain the following
    GPE data:

    0. the mean motion (in [rad / min]),
    1. the eccentricity,
    2. the inclination (in [rad]),
    3. the right ascension of the ascending node (in [rad]),
    4. the argument of perigee (in [rad]),
    5. the mean anomaly (in [rad]),
    6. the `BSTAR <https://en.wikipedia.org/wiki/BSTAR>`__ drag term (in the same unit as given in the GPE),
    7. the reference epoch (as a Julian date),
    8. a fractional correction to the epoch (in Julian days).

    Epochs must be provided in the UTC scale of time, following the convention that days in which
    leap seconds are added/removed are 1 second longer/shorter than one SI day.

    The *diff_order* argument indicates the desired differentiation order. If equal to 0, then
    derivatives are disabled.

    The :ref:`fp_type keyword argument <api_common_kwargs_fp_type>` can be passed in *kwargs*
    to select the precision of the propagator (double-precision is the default, single-precision
    is also supported).

    *kwargs* can also optionally contain keyword arguments from the :ref:`api_common_kwargs_llvm` set
    and the :ref:`api_common_kwargs_cfunc` set.

    :param sat_list: the GPE data.
    :param diff_order: the derivatives order.

    :raises TypeError: if an unsupported :ref:`fp_type <api_common_kwargs_fp_type>` is specified.
    :raises: any exception raised by the construction of the propagator.

    :returns: an SGP4 propagator object.
    :rtype: sgp4_propagator_dbl | sgp4_propagator_flt

    """
    from .. import _fp_to_suffix_dict, core

    fp_type = kwargs.pop("fp_type", float)

    try:
        fp_suffix = _fp_to_suffix_dict[fp_type]
    except KeyError:
        raise TypeError(f"Unknown fp type '{fp_type}'")

    if hasattr(core, f"_model_sgp4_propagator{fp_suffix}"):
        return getattr(core, f"_model_sgp4_propagator{fp_suffix}")(
            sat_list, diff_order, **kwargs
        )
    else:
        raise TypeError(f"No sgp4 propagator available for the fp type '{fp_type}'")


del _core, _lst, _name, numpy, Union



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.


import unittest as _ut


class cfunc_test_case(_ut.TestCase):
    def test_basic(self):
        from . import cfunc, make_vars, core, par, time, code_model
        import pickle
        from copy import copy, deepcopy
        from sys import getrefcount

        self.assertRaises(ValueError, lambda: cfunc([], []))

        x, y, z, s = make_vars("x", "y", "z", "s")
        cf = cfunc([y * (x + z)], [x, y, z])

        # Ensure that when extracting the llvm states
        # the ref count of cf is properly updated.
        rc = getrefcount(cf)
        _tmp = cf.llvm_states
        self.assertEqual(getrefcount(cf), rc + 3)

        self.assertTrue(all(not _.force_avx512 for _ in cf.llvm_states))
        self.assertTrue(all(not _.slp_vectorize for _ in cf.llvm_states))

        self.assertEqual(cf([1, 2, 3]), copy(cf)([1, 2, 3]))
        self.assertEqual(cf([1, 2, 3]), deepcopy(cf)([1, 2, 3]))
        self.assertEqual(cf([1, 2, 3]), pickle.loads(pickle.dumps(cf))([1, 2, 3]))

        self.assertEqual(cf.vars, [x, y, z])
        self.assertEqual(cf.fn, [y * (x + z)])
        self.assertEqual(len(cf.dc), 6)
        self.assertTrue(all(len(_.ir) != 0 for _ in cf.llvm_states))
        self.assertEqual(deepcopy(cf).vars, [x, y, z])
        self.assertEqual(deepcopy(cf).fn, [y * (x + z)])
        self.assertEqual(deepcopy(cf).dc, cf.dc)
        self.assertEqual(
            [_.ir for _ in deepcopy(cf).llvm_states],
            [_.ir for _ in cf.llvm_states],
        )
        self.assertEqual(pickle.loads(pickle.dumps(cf)).vars, [x, y, z])
        self.assertEqual(pickle.loads(pickle.dumps(cf)).fn, [y * (x + z)])
        self.assertEqual(pickle.loads(pickle.dumps(cf)).dc, cf.dc)
        self.assertEqual(
            [_.ir for _ in pickle.loads(pickle.dumps(cf)).llvm_states],
            [_.ir for _ in cf.llvm_states],
        )

        # Construct with custom LLVM settings too.
        cf = cfunc(
            [y * (x + z)],
            vars=[y, z, x],
            force_avx512=True,
            slp_vectorize=True,
            compact_mode=True,
            parjit=False,
            code_model=code_model.large,
        )

        rc = getrefcount(cf)
        _tmp = cf.llvm_states
        self.assertEqual(getrefcount(cf), rc + 1)

        self.assertEqual(cf.vars, [y, z, x])
        self.assertTrue(cf.llvm_states.force_avx512)
        self.assertTrue(cf.llvm_states.slp_vectorize)
        self.assertFalse(cf.llvm_states.parjit)
        self.assertEqual(cf.llvm_states.code_model, code_model.large)

        # Tests for correct detection of number of params, time dependency
        # and list of variables.
        cf = cfunc([y * (x + z), x], vars=[y, z, x])
        self.assertEqual(cf.nparams, 0)
        cf = cfunc([y * (x + z), par[0]], vars=[y, z, x])
        self.assertEqual(cf.nparams, 1)
        cf = cfunc([y * (x + z) - par[89], par[0]], vars=[y, z, x])
        self.assertEqual(cf.nparams, 90)

        cf = cfunc([y * (x + z), x], vars=[y, z, x])
        self.assertFalse(cf.is_time_dependent)
        cf = cfunc([y * (x + z) + time, x], vars=[y, z, x])
        self.assertTrue(cf.is_time_dependent)
        cf = cfunc([y * (x + z), x + time], vars=[y, z, x])
        self.assertTrue(cf.is_time_dependent)

        cf = cfunc([y * (x + z), x + time], [x, y, z])
        self.assertEqual(cf.vars, [x, y, z])
        cf = cfunc([y * (x + z), x + time], vars=[y, z, x])
        self.assertEqual(cf.vars, [y, z, x])
        cf = cfunc([y * (x + z), x + time], vars=[y, z, x, s])
        self.assertEqual(cf.vars, [y, z, x, s])
        cf = cfunc([y * (x + z), x + time], vars=[s, y, z, x])
        self.assertEqual(cf.vars, [s, y, z, x])

        # NOTE: test for a bug in the multiprecision
        # implementation where the precision is not
        # correctly copied.
        if not hasattr(core, "real"):
            return

        real = core.real

        cf = cfunc([y * (x + z)], [x, y, z], fp_type=real, prec=128)
        self.assertEqual(
            cf([real(1, 128), real(2, 128), real(3, 128)]),
            copy(cf)([real(1, 128), real(2, 128), real(3, 128)]),
        )

        self.assertEqual(cf.prec, 128)
        self.assertEqual(deepcopy(cf).prec, 128)
        self.assertEqual(pickle.loads(pickle.dumps(cf)).prec, 128)

    def test_multi(self):
        import numpy as np
        from . import cfunc, make_vars, sin, par, expression, core, time
        from .core import _ppc_arch
        from .test import _get_eps, _allclose

        if _ppc_arch:
            fp_types = [np.float32, float]
        else:
            fp_types = [np.float32, float, np.longdouble]

        if hasattr(core, "real128"):
            fp_types.append(core.real128)

        x, y = make_vars("x", "y")
        func = [sin(x + y), x - par[0], x + y + par[1] * time]

        for fp_t in fp_types:
            fn = cfunc(func, vars=[y, x], fp_type=fp_t, batch_size=0)
            self.assertTrue(fn.batch_size > 0)

            if fp_t == np.longdouble:
                with self.assertRaises(ValueError) as cm:
                    cfunc(func, vars=[y, x], fp_type=fp_t, batch_size=2)
                self.assertTrue(
                    "Batch sizes greater than 1 are not supported for this"
                    " floating-point type" in str(cm.exception)
                )

            fn = cfunc(func, vars=[y, x], fp_type=fp_t)

            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2, 5), dtype=fp_t),
                    pars=[fp_t(0)],
                    outputs=np.zeros((3, 1), dtype=fp_t),
                    time=np.zeros((5,), dtype=fp_t),
                )
            self.assertTrue(
                "The array of parameters provided for the evaluation of a compiled"
                " function has 1 dimension(s), but it must have 2 dimension(s) instead"
                " (i.e., the same number of dimensions as the array of inputs)"
                in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2, 5), dtype=fp_t),
                    pars=np.zeros(
                        (2, 4),
                        dtype=fp_t,
                    ),
                    time=np.zeros((5,), dtype=fp_t),
                )
            self.assertTrue(
                "The array of parameter values provided for the evaluation of a"
                " compiled function has 4 column(s), but the expected number of columns"
                " deduced from the outputs array is 5" in str(cm.exception)
            )

            nw_arr = np.zeros((3, 5), dtype=fp_t)
            nw_arr.setflags(write=False)
            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2, 5), dtype=fp_t),
                    outputs=nw_arr,
                    pars=np.zeros((2, 5), dtype=fp_t),
                    time=np.zeros((5,), dtype=fp_t),
                )
            self.assertTrue(
                "The array of outputs provided for the evaluation of a compiled"
                " function is not writeable" in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2, 5), dtype=fp_t),
                    pars=np.zeros((2, 5), dtype=fp_t),
                )
            self.assertTrue(
                "An array of time values must be provided in order to evaluate a"
                " time-dependent function" in str(cm.exception)
            )

            with self.assertRaises(TypeError) as cm:
                fn(
                    np.zeros((2,), dtype=fp_t),
                    pars=np.zeros((2,), dtype=fp_t),
                    time=np.zeros((5,), dtype=fp_t),
                )
            self.assertTrue(
                "The time value cannot be an array when evaluating a compiled function"
                " over a single set of inputs, it should be a scalar instead"
                in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2, 5), dtype=fp_t),
                    pars=np.zeros((2, 5), dtype=fp_t),
                    time=np.zeros((5, 5), dtype=fp_t),
                )
            self.assertTrue(
                "The array of times provided for the evaluation of a compiled function"
                " has 2 dimension(s), but it must be one-dimensional instead"
                in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2, 5), dtype=fp_t),
                    pars=np.zeros((2, 5), dtype=fp_t),
                    time=np.zeros((6,), dtype=fp_t),
                )
            self.assertTrue(
                "The array of time values provided for the evaluation of a compiled"
                " function has a size of 6, but the expected size deduced from the"
                " outputs array is 5" in str(cm.exception)
            )

            # Check with wrong dtypes.
            if fp_t != float:
                with self.assertRaises(TypeError) as cm:
                    fn(
                        np.zeros((2, 5), dtype=float),
                        pars=np.zeros((2, 5), dtype=fp_t),
                        time=np.zeros((1,), dtype=fp_t),
                    )
                self.assertTrue(
                    "Invalid dtype detected for the inputs of a compiled function: the"
                    " expected dtype " in str(cm.exception)
                )

                with self.assertRaises(TypeError) as cm:
                    fn(
                        np.zeros((2, 5), dtype=fp_t),
                        pars=np.zeros((2, 5), dtype=float),
                        time=np.zeros((1,), dtype=fp_t),
                    )
                self.assertTrue(
                    "Invalid dtype detected for the parameters of a compiled function:"
                    " the expected dtype " in str(cm.exception)
                )

                with self.assertRaises(TypeError) as cm:
                    fn(
                        np.zeros((2, 5), dtype=fp_t),
                        pars=np.zeros((2, 5), dtype=fp_t),
                        time=np.zeros((1,), dtype=fp_t),
                        outputs=np.zeros((3, 5), dtype=float),
                    )
                self.assertTrue(
                    "Invalid dtype detected for the outputs of a compiled function: the"
                    " expected dtype " in str(cm.exception)
                )

                with self.assertRaises(TypeError) as cm:
                    fn(
                        np.zeros((2, 5), dtype=fp_t),
                        pars=np.zeros((2, 5), dtype=fp_t),
                        time=np.zeros((1,), dtype=float),
                    )
                self.assertTrue(
                    "Invalid dtype detected for the time values of a compiled function:"
                    " the expected dtype " in str(cm.exception)
                )

            # Non-contiguous time array.
            tarr = np.zeros((10,), dtype=fp_t)
            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2, 5), dtype=fp_t),
                    pars=np.zeros((2, 5), dtype=fp_t),
                    time=tarr[::2],
                )
            self.assertTrue(
                "Invalid time array detected: the array is not C-style contiguous,"
                " please consider using numpy.ascontiguousarray() to turn it into one"
                in str(cm.exception)
            )

            for nevals in range(0, 10):
                fn = cfunc(func, vars=[y, x], fp_type=fp_t)

                # NOTE: deterministic seeding.
                rng = np.random.default_rng(nevals)

                # NOTE: long double rng not supported.
                inputs = rng.random((2, nevals), dtype=float).astype(fp_t)
                pars = rng.random((2, nevals), dtype=float).astype(fp_t)
                tm = rng.random((nevals,), dtype=float).astype(fp_t)

                eval_arr = fn(inputs=inputs, pars=pars, time=tm)
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        np.sin(inputs[1, :] + inputs[0, :]),
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        inputs[1, :] - pars[0, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[2],
                        inputs[1, :] + inputs[0, :] + pars[1, :] * tm,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                # A couple of tests with the batch_parallel flag.
                eval_arr2 = fn(inputs=inputs, pars=pars, time=tm, batch_parallel=True)
                self.assertTrue(np.all(eval_arr == eval_arr2))

                eval_arr2 = fn(inputs=inputs, pars=pars, time=tm, batch_parallel=False)
                self.assertTrue(np.all(eval_arr == eval_arr2))

                eval_arr2 = fn(inputs=inputs, pars=pars, time=tm, batch_parallel=None)
                self.assertTrue(np.all(eval_arr == eval_arr2))

                # NOTE: simple function without pars used to test
                # that the null pars span is created with the correct shape
                # in a parameter-less function.
                simple_fn = cfunc([x + y], vars=[x, y], fp_type=fp_t)
                eval_arr = simple_fn(inputs=inputs)
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        inputs[0, :] + inputs[1, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                # Check that eval_arr actually uses the memory
                # provided from the outputs argument.
                out_arr = np.zeros((3, nevals), dtype=fp_t)
                eval_arr = fn(inputs=inputs, pars=pars, outputs=out_arr, time=tm)
                self.assertEqual(id(eval_arr), id(out_arr))
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        np.sin(inputs[1, :] + inputs[0, :]),
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        inputs[1, :] - pars[0, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[2],
                        inputs[1, :] + inputs[0, :] + pars[1, :] * tm,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                # Test with non-owning arrays.
                eval_arr = fn(inputs=inputs[:], pars=pars, time=tm)
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        np.sin(inputs[1, :] + inputs[0, :]),
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        inputs[1, :] - pars[0, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[2],
                        inputs[1, :] + inputs[0, :] + pars[1, :] * tm,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                eval_arr = fn(inputs=inputs, pars=pars[:], time=tm)
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        np.sin(inputs[1, :] + inputs[0, :]),
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        inputs[1, :] - pars[0, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[2],
                        inputs[1, :] + inputs[0, :] + pars[1, :] * tm,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                eval_arr = fn(inputs=inputs, pars=pars, time=tm[:])
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        np.sin(inputs[1, :] + inputs[0, :]),
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        inputs[1, :] - pars[0, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[2],
                        inputs[1, :] + inputs[0, :] + pars[1, :] * tm,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                # Test with non-distinct arrays.
                if nevals == 0:
                    eval_arr = fn(inputs=inputs, pars=inputs, time=tm)
                    self.assertTrue(
                        _allclose(
                            eval_arr[0],
                            np.sin(inputs[1, :] + inputs[0, :]),
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                    self.assertTrue(
                        _allclose(
                            eval_arr[1],
                            inputs[1, :] - inputs[0, :],
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                    self.assertTrue(
                        _allclose(
                            eval_arr[2],
                            inputs[1, :] + inputs[0, :] + inputs[1, :] * tm,
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )

                    eval_arr = fn(inputs=inputs, pars=pars, time=pars[0, :])
                    self.assertTrue(
                        _allclose(
                            eval_arr[0],
                            np.sin(inputs[1, :] + inputs[0, :]),
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                    self.assertTrue(
                        _allclose(
                            eval_arr[1],
                            inputs[1, :] - pars[0, :],
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                    self.assertTrue(
                        _allclose(
                            eval_arr[2],
                            inputs[1, :] + inputs[0, :] + pars[1, :] * pars[0, :],
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                else:
                    with self.assertRaises(ValueError) as cm:
                        fn(inputs=inputs, pars=inputs, time=tm)
                    self.assertTrue(
                        "Potential memory overlaps detected when attempting to evaluate"
                        " a compiled function: please make sure that all input arrays"
                        " are distinct" in str(cm.exception)
                    )
                    with self.assertRaises(ValueError) as cm:
                        fn(inputs=inputs, pars=pars, time=pars[0, :])
                    self.assertTrue(
                        "Potential memory overlaps detected when attempting to evaluate"
                        " a compiled function: please make sure that all input arrays"
                        " are distinct" in str(cm.exception)
                    )

                # Test with arrays which are not C style.
                if nevals == 0:
                    inputs = rng.random((4, nevals), dtype=float).astype(fp_t)
                    inputs = inputs[::2]
                    eval_arr = fn(inputs=inputs, pars=pars, time=tm)
                    self.assertTrue(
                        _allclose(
                            eval_arr[0],
                            np.sin(inputs[1, :] + inputs[0, :]),
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                    self.assertTrue(
                        _allclose(
                            eval_arr[1],
                            inputs[1, :] - pars[0, :],
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                    self.assertTrue(
                        _allclose(
                            eval_arr[2],
                            inputs[1, :] + inputs[0, :] + pars[1, :] * tm,
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )

                    pars = rng.random((4, nevals), dtype=float).astype(fp_t)
                    pars = pars[::2]
                    eval_arr = fn(inputs=inputs, pars=pars, time=tm)
                    self.assertTrue(
                        _allclose(
                            eval_arr[0],
                            np.sin(inputs[1, :] + inputs[0, :]),
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                    self.assertTrue(
                        _allclose(
                            eval_arr[1],
                            inputs[1, :] - pars[0, :],
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                    self.assertTrue(
                        _allclose(
                            eval_arr[2],
                            inputs[1, :] + inputs[0, :] + pars[1, :] * tm,
                            rtol=_get_eps(fp_t) * 10,
                            atol=_get_eps(fp_t) * 10,
                        )
                    )
                else:
                    inputs = rng.random((4, nevals), dtype=float).astype(fp_t)
                    inputs = inputs[::2]
                    with self.assertRaises(ValueError) as cm:
                        fn(inputs=inputs, pars=pars, time=tm)
                    self.assertTrue(
                        "Invalid inputs array detected: the array is not C-style"
                        " contiguous, please consider using numpy.ascontiguousarray()"
                        " to turn it into one" in str(cm.exception)
                    )

                    pars = rng.random((4, nevals), dtype=float).astype(fp_t)
                    pars = pars[::2]
                    with self.assertRaises(ValueError) as cm:
                        fn(inputs=inputs, pars=pars, time=tm)
                    self.assertTrue(
                        "Invalid inputs array detected: the array is not C-style"
                        " contiguous, please consider using numpy.ascontiguousarray()"
                        " to turn it into one" in str(cm.exception)
                    )

                # Tests with no inputs.
                fn = cfunc(
                    [expression(fp_t(3)) + par[1], par[0] + time], [], fp_type=fp_t
                )

                inputs = rng.random((0, nevals), dtype=float).astype(fp_t)
                pars = rng.random((2, nevals), dtype=float).astype(fp_t)
                eval_arr = fn(inputs=inputs, pars=pars, time=tm)
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        3 + pars[1, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        pars[0, :] + tm,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                eval_arr = fn(inputs=inputs[:], pars=pars, time=tm)
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        3 + pars[1, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        pars[0, :] + tm,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                eval_arr = fn(inputs=inputs, pars=pars[:], time=tm)
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        3 + pars[1, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        pars[0, :] + tm,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                eval_arr = fn(inputs=inputs, pars=pars, time=tm[:])
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        3 + pars[1, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        pars[0, :] + tm,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                fn = cfunc([expression(fp_t(3)), expression(fp_t(4))], [], fp_type=fp_t)

                inputs = rng.random((0, nevals), dtype=float).astype(fp_t)
                pars = rng.random((0, nevals), dtype=float).astype(fp_t)
                eval_arr = fn(inputs=inputs, pars=pars)
                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        [3] * nevals,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )
                self.assertTrue(
                    _allclose(
                        eval_arr[1],
                        [4] * nevals,
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                # Test case in which there are no pars but a pars array is provided anyway,
                # with the correct shape.
                fn = cfunc([x + y], [x, y], fp_type=fp_t)
                inputs = rng.random((2, nevals), dtype=float).astype(fp_t)
                eval_arr = fn(inputs=inputs, pars=np.zeros((0, nevals), dtype=fp_t))

                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        inputs[0, :] + inputs[1, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

                # Test a case without time but a time array is provided anyway, with
                # the correct shape.
                eval_arr = fn(
                    inputs=inputs, pars=np.zeros((0, nevals), dtype=fp_t), time=tm
                )

                self.assertTrue(
                    _allclose(
                        eval_arr[0],
                        inputs[0, :] + inputs[1, :],
                        rtol=_get_eps(fp_t) * 10,
                        atol=_get_eps(fp_t) * 10,
                    )
                )

        # Check throwing behaviour with long double on PPC.
        if _ppc_arch:
            with self.assertRaises(NotImplementedError):
                cfunc(func, vars=[y, x], fp_type=np.longdouble)

    def test_single(self):
        import numpy as np
        from . import cfunc, make_vars, sin, par, expression, core, time
        from .core import _ppc_arch
        from .test import _get_eps, _allclose

        if _ppc_arch:
            fp_types = [np.float32, float]
        else:
            fp_types = [np.float32, float, np.longdouble]

        if hasattr(core, "real128"):
            fp_types.append(core.real128)

        x, y = make_vars("x", "y")
        func = [sin(x + y), x - par[0], x + y + par[1] + time]

        # NOTE: perhaps in the future we can add
        # some more testing for high_accuracy, compact_mode,
        # etc., once we figure out how to test for them. Perhaps
        # examine the llvm states?
        for fp_t in fp_types:
            fn = cfunc(func, [x, y], fp_type=fp_t)

            with self.assertRaises(ValueError) as cm:
                fn([fp_t(1), fp_t(2)])
            self.assertTrue(
                "The array of parameter values provided for the evaluation of a"
                " compiled function has 0 element(s), but the number of parameters in"
                " the function is 2" in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn(np.zeros((1, 2, 3), dtype=fp_t), pars=[fp_t(0)], time=fp_t(0))
            self.assertTrue(
                "The array of inputs provided for the evaluation of a compiled function"
                " has 3 dimensions, but it must have either 1 or 2 dimensions instead"
                in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn([fp_t(0)], pars=[fp_t(0)], time=fp_t(0))
            self.assertTrue(
                "Invalid inputs array passed to a cfunc: the number of function inputs"
                " is 2, but the inputs array has a size of 1" in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2,), dtype=fp_t),
                    pars=[fp_t(0)],
                    outputs=np.zeros((2,), dtype=fp_t),
                    time=fp_t(0),
                )
            self.assertTrue(
                "Invalid outputs array passed to a cfunc: the number of function"
                " outputs is 3, but the outputs array has a size of 2"
                in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2,), dtype=fp_t),
                    pars=np.zeros((0,), dtype=fp_t),
                    time=fp_t(0),
                )
            self.assertTrue(
                "The array of parameter values provided for the evaluation of a"
                " compiled function has 0 element(s), but the number of parameters in"
                " the function is 2" in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn(
                    np.zeros((2,), dtype=fp_t),
                    pars=np.zeros((2,), dtype=fp_t),
                )
            self.assertTrue(
                "A time value must be provided in order to evaluate a time-dependent"
                " function" in str(cm.exception)
            )

            eval_arr = fn([fp_t(1), fp_t(2)], pars=[fp_t(-5), fp_t(1)], time=fp_t(3))
            self.assertTrue(
                _allclose(
                    eval_arr,
                    [
                        np.sin(fp_t(1) + fp_t(2)),
                        fp_t(1) - fp_t(-5),
                        fp_t(1) + fp_t(2) + fp_t(1) + fp_t(3),
                    ],
                    rtol=_get_eps(fp_t) * 10,
                    atol=_get_eps(fp_t) * 10,
                )
            )

            # Check that eval_arr actually uses the memory
            # provided from the outputs argument.
            out_arr = np.zeros((3,), dtype=fp_t)
            eval_arr = fn(
                [fp_t(1), fp_t(2)],
                pars=[fp_t(-5), fp_t(1)],
                outputs=out_arr,
                time=fp_t(3),
            )
            self.assertTrue(np.shares_memory(eval_arr, out_arr))
            self.assertEqual(id(eval_arr), id(out_arr))
            self.assertTrue(
                _allclose(
                    eval_arr,
                    [
                        np.sin(fp_t(1) + fp_t(2)),
                        fp_t(1) - fp_t(-5),
                        fp_t(1) + fp_t(2) + fp_t(1) + fp_t(3),
                    ],
                    rtol=_get_eps(fp_t) * 10,
                    atol=_get_eps(fp_t) * 10,
                )
            )

            # Test with non-owning arrays.
            inputs = np.array([fp_t(1), fp_t(2)])
            eval_arr = fn(inputs=inputs[:], pars=[fp_t(-5), fp_t(1)], time=fp_t(3))
            self.assertTrue(
                _allclose(
                    eval_arr,
                    [
                        np.sin(fp_t(1) + fp_t(2)),
                        fp_t(1) - fp_t(-5),
                        fp_t(1) + fp_t(2) + fp_t(1) + fp_t(3),
                    ],
                    rtol=_get_eps(fp_t) * 10,
                    atol=_get_eps(fp_t) * 10,
                )
            )

            pars = np.array([fp_t(-5), fp_t(1)])
            eval_arr = fn(inputs=inputs, pars=pars[:], time=fp_t(3))
            self.assertTrue(
                _allclose(
                    eval_arr,
                    [
                        np.sin(fp_t(1) + fp_t(2)),
                        fp_t(1) - fp_t(-5),
                        fp_t(1) + fp_t(2) + fp_t(1) + fp_t(3),
                    ],
                    rtol=_get_eps(fp_t) * 10,
                    atol=_get_eps(fp_t) * 10,
                )
            )

            # Test with non-distinct arrays.
            with self.assertRaises(ValueError) as cm:
                fn(inputs=inputs, pars=inputs, time=fp_t(3))
            self.assertTrue(
                "Potential memory overlaps detected when attempting to evaluate a"
                " compiled function: please make sure that all input arrays are"
                " distinct" in str(cm.exception)
            )

            with self.assertRaises(ValueError) as cm:
                fn(inputs=inputs, pars=pars, time=fp_t(3), outputs=inputs)
            self.assertTrue(
                "Potential memory overlaps detected when attempting to evaluate a"
                " compiled function: please make sure that all input arrays are"
                " distinct" in str(cm.exception)
            )

            # Test with arrays which are not C style.
            inputs = np.zeros((4,), dtype=fp_t)
            inputs[::2] = [fp_t(1), fp_t(2)]
            with self.assertRaises(ValueError) as cm:
                fn(inputs=inputs[::2], pars=pars, time=fp_t(3))
            self.assertTrue(
                "Invalid inputs array detected: the array is not C-style contiguous,"
                " please consider using numpy.ascontiguousarray() to turn it into one"
                in str(cm.exception)
            )

            inputs = np.array([fp_t(1), fp_t(2)])
            pars = np.zeros((4,), dtype=fp_t)
            pars[::2] = [fp_t(-5), fp_t(1)]
            with self.assertRaises(ValueError) as cm:
                fn(inputs=inputs, pars=pars[::2], time=fp_t(3))
            self.assertTrue(
                "Invalid parameters array detected: the array is not C-style"
                " contiguous, please consider using numpy.ascontiguousarray() to turn"
                " it into one" in str(cm.exception)
            )

            inputs = np.array([fp_t(1), fp_t(2)])
            pars = np.zeros((4,), dtype=fp_t)
            outputs = np.zeros((6,), dtype=fp_t)
            with self.assertRaises(ValueError) as cm:
                fn(inputs=inputs, pars=pars, outputs=outputs[::2], time=fp_t(3))
            self.assertTrue(
                "Invalid outputs array detected: the array is not C-style contiguous,"
                " please consider using numpy.ascontiguousarray() to turn it into one"
                in str(cm.exception)
            )

            # Check with wrong dtypes.
            if fp_t != float:
                with self.assertRaises(TypeError) as cm:
                    fn(
                        np.zeros((2,), dtype=float),
                        pars=np.zeros((2,), dtype=fp_t),
                        time=fp_t(),
                    )
                self.assertTrue(
                    "Invalid dtype detected for the inputs of a compiled function: the"
                    " expected dtype " in str(cm.exception)
                )

                with self.assertRaises(TypeError) as cm:
                    fn(
                        np.zeros((2,), dtype=fp_t),
                        pars=np.zeros((2,), dtype=float),
                        time=fp_t(),
                    )
                self.assertTrue(
                    "Invalid dtype detected for the parameters of a compiled function:"
                    " the expected dtype " in str(cm.exception)
                )

                with self.assertRaises(TypeError) as cm:
                    fn(
                        np.zeros((2,), dtype=fp_t),
                        pars=np.zeros((2,), dtype=fp_t),
                        time=fp_t(),
                        outputs=np.zeros((3,), dtype=float),
                    )
                self.assertTrue(
                    "Invalid dtype detected for the outputs of a compiled function: the"
                    " expected dtype " in str(cm.exception)
                )

            # Time as an array instead of scalar.
            with self.assertRaises(TypeError) as cm:
                fn(
                    np.zeros((2,), dtype=fp_t),
                    pars=np.zeros((2,), dtype=fp_t),
                    time=np.zeros((1,), dtype=fp_t),
                )
            self.assertTrue(
                "The time value cannot be an array when evaluating a compiled function"
                " over a single set of inputs, it should be a scalar instead"
                in str(cm.exception)
            )

            # Tests with no inputs.
            fn = cfunc([expression(fp_t(3)) + par[1], par[0] + time], [], fp_type=fp_t)

            eval_arr = fn(
                inputs=np.zeros((0,), dtype=fp_t), pars=[fp_t(1), fp_t(2)], time=fp_t(3)
            )
            self.assertTrue(
                _allclose(
                    eval_arr,
                    [fp_t(3) + 2, fp_t(1) + fp_t(3)],
                    rtol=_get_eps(fp_t) * 10,
                    atol=_get_eps(fp_t) * 10,
                )
            )

            inputs = np.zeros((0,), dtype=fp_t)
            eval_arr = fn(inputs=inputs[:], pars=[fp_t(1), fp_t(2)], time=fp_t(3))
            self.assertTrue(
                _allclose(
                    eval_arr,
                    [fp_t(3) + 2, fp_t(1) + fp_t(3)],
                    rtol=_get_eps(fp_t) * 10,
                    atol=_get_eps(fp_t) * 10,
                )
            )

            fn = cfunc(
                [expression(fp_t(3)), expression(fp_t(4)) + time], [], fp_type=fp_t
            )

            eval_arr = fn(
                inputs=np.zeros((0,), dtype=fp_t),
                pars=np.zeros((0,), dtype=fp_t),
                time=fp_t(3),
            )
            self.assertTrue(
                _allclose(
                    eval_arr,
                    [fp_t(3), fp_t(4) + fp_t(3)],
                    rtol=_get_eps(fp_t) * 10,
                    atol=_get_eps(fp_t) * 10,
                )
            )

            eval_arr = fn(
                inputs=np.zeros((0,), dtype=fp_t),
                pars=np.zeros((0,), dtype=fp_t),
                time=fp_t(3),
            )
            self.assertTrue(
                _allclose(
                    eval_arr,
                    [fp_t(3), fp_t(4) + fp_t(3)],
                    rtol=_get_eps(fp_t) * 10,
                    atol=_get_eps(fp_t) * 10,
                )
            )

            # Test case in which there are no pars but a pars array is provided anyway,
            # with the correct shape.
            fn = cfunc([x + y], [x, y], fp_type=fp_t)
            eval_arr = fn(inputs=[fp_t(1), fp_t(2)], pars=np.zeros((0,), dtype=fp_t))

            self.assertEqual(eval_arr[0], 3)

            # Same but with time.
            eval_arr = fn(
                inputs=[fp_t(1), fp_t(2)], pars=np.zeros((0,), dtype=fp_t), time=fp_t(3)
            )

            self.assertEqual(eval_arr[0], 3)



# Copyright 2020-2025 Francesco Biscani (bluescarni@gmail.com), Dario Izzo (dario.izzo@gmail.com)
#
# This file is part of the heyoka.py library.
#
# This Source Code Form is subject to the terms of the Mozilla
# Public License v. 2.0. If a copy of the MPL was not distributed
# with this file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest as _ut


class expression_test_case(_ut.TestCase):
    def test_basic(self):
        from . import expression as ex, core
        from . import make_vars
        import numpy as np

        with_real128 = hasattr(core, "real128")
        ld_63bit = np.finfo(np.longdouble).nmant == 63

        if with_real128:
            real128 = core.real128

        # Minimal make_vars() testing.
        with self.assertRaises(ValueError) as cm:
            make_vars()
        self.assertTrue(
            "At least one argument is required when invoking 'make_vars()'"
            in str(cm.exception)
        )

        x = make_vars("x")
        self.assertTrue(isinstance(x, ex))
        self.assertEqual(x, ex("x"))

        l = make_vars("x", "y")
        self.assertTrue(isinstance(l, list))
        x, y = l
        self.assertEqual(x, ex("x"))
        self.assertEqual(y, ex("y"))

        # Constructors.
        self.assertEqual(ex(), ex(0))
        self.assertEqual(ex(123), ex(123.0))
        self.assertEqual(str(ex(123)), "123.00000000000000")
        self.assertEqual(str(ex(np.float32("1.1"))), "1.10000002")

        # Error with large integer.
        with self.assertRaises(TypeError) as cm:
            ex(123 << 56)
        self.assertTrue("incompatible constructor arguments" in str(cm.exception))

        self.assertEqual(str(ex(1.1)), "1.1000000000000001")

        if ld_63bit:
            self.assertEqual(str(ex(np.longdouble("1.1"))), "1.10000000000000000002")

        if with_real128:
            self.assertEqual(
                str(ex(real128("1.1"))), "1.10000000000000000000000000000000008"
            )

        self.assertEqual(str(ex("x")), "x")

        # Unary arithmetics.
        self.assertEqual(ex(42), +ex(42))
        self.assertEqual(ex(-42), -ex(42))

        # Addition.
        self.assertEqual(ex(42) + ex(-1), ex(41))
        self.assertEqual(ex(42) + -1, ex(41))
        self.assertEqual(-1 + ex(42), ex(41))
        with self.assertRaises(TypeError) as cm:
            ex(42) + (2 << 112)
        with self.assertRaises(TypeError) as cm:
            (2 << 112) + ex(42)
        self.assertEqual(ex(42) + -1.1, ex(40.899999999999999))
        self.assertEqual(-1.1 + ex(42), ex(40.899999999999999))
        if ld_63bit:
            self.assertEqual(
                ex(42) + np.longdouble("-1.1"),
                ex(np.longdouble("40.9000000000000000014")),
            )
            self.assertEqual(
                np.longdouble("-1.1") + ex(42),
                ex(np.longdouble("40.9000000000000000014")),
            )
        if with_real128:
            self.assertEqual(
                ex(42) + real128("-1.1"),
                ex(real128("40.8999999999999999999999999999999988")),
            )
            self.assertEqual(
                real128("-1.1") + ex(42),
                ex(real128("40.8999999999999999999999999999999988")),
            )

        # Subtraction.
        self.assertEqual(ex(42) - ex(-1), ex(43))
        self.assertEqual(ex(42) - -1, ex(43))
        self.assertEqual(-1 - ex(42), ex(-43))
        with self.assertRaises(TypeError) as cm:
            ex(42) - (2 << 112)
        with self.assertRaises(TypeError) as cm:
            (2 << 112) - ex(42)
        self.assertEqual(ex(42) - -1.1, ex(43.100000000000001))
        self.assertEqual(-1.1 - ex(42), ex(-43.100000000000001))
        if ld_63bit:
            self.assertEqual(
                ex(42) - np.longdouble("-1.1"),
                ex(np.longdouble("43.0999999999999999986")),
            )
            self.assertEqual(
                np.longdouble("-1.1") - ex(42),
                ex(np.longdouble("-43.0999999999999999986")),
            )
        if with_real128:
            self.assertEqual(
                ex(42) - real128("-1.1"),
                ex(real128("43.1000000000000000000000000000000012")),
            )
            self.assertEqual(
                real128("-1.1") - ex(42),
                ex(real128("-43.1000000000000000000000000000000012")),
            )

        # Multiplication.
        self.assertEqual(ex(42) * ex(-1), ex(-42))
        self.assertEqual(ex(42) * -1, ex(-42))
        self.assertEqual(-1 * ex(42), ex(-42))
        with self.assertRaises(TypeError) as cm:
            ex(42) * (2 << 112)
        with self.assertRaises(TypeError) as cm:
            (2 << 112) * ex(42)
        self.assertEqual(ex(42) * -1.1, ex(-46.200000000000003))
        self.assertEqual(-1.1 * ex(42), ex(-46.200000000000003))
        if ld_63bit:
            self.assertEqual(
                ex(42) * np.longdouble("-1.1"),
                ex(np.longdouble("-46.2000000000000000007")),
            )
            self.assertEqual(
                np.longdouble("-1.1") * ex(42),
                ex(np.longdouble("-46.2000000000000000007")),
            )
        if with_real128:
            self.assertEqual(
                ex(42) * real128("-1.1"),
                ex(real128("-46.2000000000000000000000000000000025")),
            )
            self.assertEqual(
                real128("-1.1") * ex(42),
                ex(real128("-46.2000000000000000000000000000000025")),
            )

        # Division.
        self.assertEqual(ex(42) / ex(-1), ex(-42))
        self.assertEqual(ex(42) / -1, ex(-42))
        self.assertEqual(-42 / ex(1), ex(-42))
        with self.assertRaises(TypeError) as cm:
            ex(42) / (2 << 112)
        with self.assertRaises(TypeError) as cm:
            (2 << 112) / ex(42)
        self.assertEqual(ex(42) / -1.1, ex(-38.181818181818180))
        self.assertEqual(-1.1 / ex(42), ex(-0.02619047619047619))
        if ld_63bit:
            self.assertEqual(
                ex(42) / np.longdouble("-1.1"),
                ex(np.longdouble("-38.1818181818181818163")),
            )
            self.assertEqual(
                np.longdouble("-1.1") / ex(42),
                ex(np.longdouble("-0.0261904761904761904772")),
            )
        if with_real128:
            self.assertEqual(
                ex(42) / real128("-1.1"),
                ex(real128("-38.1818181818181818181818181818181801")),
            )
            self.assertEqual(
                real128("-1.1") / ex(42),
                ex(real128("-0.0261904761904761904761904761904761910")),
            )

        # Comparison.
        self.assertEqual(ex(42), ex(42))
        self.assertEqual(ex("x"), ex("x"))
        self.assertNotEqual(ex(41), ex(42))
        self.assertNotEqual(ex("x"), ex("y"))

        # Exponentiation.
        self.assertEqual(str(ex("x") ** ex("y")), "x**y")
        self.assertEqual(str(ex("x") ** ex(2)), "x**2.0000000000000000")
        self.assertEqual(str(ex("x") ** ex(1.1)), "x**1.1000000000000001")
        with self.assertRaises(TypeError) as cm:
            ex(42) ** (2 << 112)
        if ld_63bit:
            self.assertEqual(
                ex(42) / np.longdouble("-1.1"),
                ex(np.longdouble("-38.1818181818181818163")),
            )
        if ld_63bit:
            self.assertEqual(
                str(ex("x") ** ex(np.longdouble("1.1"))),
                "x**1.10000000000000000002",
            )
        if with_real128:
            self.assertEqual(
                str(ex("x") ** ex(real128("1.1"))),
                "x**1.10000000000000000000000000000000008",
            )

        # Copy and deepcopy.
        from copy import copy, deepcopy

        tmp = ex("x") + ex("y")
        tmp.foo = [1, 2, 3]
        id_foo = id(tmp.foo)
        tmp_copy = copy(tmp)
        self.assertEqual(id_foo, id(tmp_copy.foo))
        tmp_dcopy = deepcopy(tmp)
        self.assertNotEqual(id_foo, id(tmp_dcopy.foo))
        self.assertEqual(tmp.foo, tmp_dcopy.foo)

    def test_copy(self):
        from . import make_vars
        from copy import copy, deepcopy

        x, y = make_vars("x", "y")
        ex = x + y

        class foo:
            pass

        ex.bar = foo()

        self.assertEqual(id(ex.bar), id(copy(ex).bar))
        self.assertNotEqual(id(ex.bar), id(deepcopy(ex).bar))
        self.assertEqual(ex, copy(ex))
        self.assertEqual(ex, deepcopy(ex))

    def test_diff(self):
        from . import make_vars, sin, cos, diff, par

        x, y = make_vars("x", "y")
        self.assertEqual(diff(cos(x * x - y), "x"), -sin(x * x - y) * (x + x))
        self.assertEqual(diff(cos(x * x - y), x), -sin(x * x - y) * (x + x))
        self.assertEqual(
            diff(cos(par[0] * par[0] - y), par[0]),
            -sin(par[0] * par[0] - y) * (par[0] + par[0]),
        )

    def test_s11n(self):
        from . import make_vars, sin, cos, core
        from .core import _ppc_arch
        from numpy import longdouble
        import pickle

        x, y = make_vars("x", "y")

        ex = x + 2.0 * y
        self.assertEqual(ex, pickle.loads(pickle.dumps(ex)))

        # Test dynamic attributes.
        ex.foo = "hello world"
        ex = pickle.loads(pickle.dumps(ex))
        self.assertEqual(ex.foo, "hello world")

        if not _ppc_arch:
            ex = sin(longdouble("1.1") * x) + 2.0 * y
            self.assertEqual(ex, pickle.loads(pickle.dumps(ex)))

        if not hasattr(core, "real128"):
            return

        from .core import real128

        # Quad precision.
        if not _ppc_arch:
            ex = sin(longdouble("1.1") * x) + real128("1.3") * cos(2.0 * y)
            self.assertEqual(ex, pickle.loads(pickle.dumps(ex)))

    def test_len(self):
        from . import make_vars

        x, y, z = make_vars("x", "y", "z")

        self.assertEqual(len(x), 1)
        self.assertEqual(len((x - y - z) + (y * z)), 13)

    def test_hash(self):
        from . import make_vars
        from copy import deepcopy

        x, y, z = make_vars("x", "y", "z")

        ex = x + y
        self.assertEqual(z * ex + ex, z * (x + y) + (x + y))
        self.assertEqual(hash(z * ex + ex), hash(z * (x + y) + (x + y)))
        ex2 = deepcopy(ex)
        self.assertEqual(hash(z * ex2 + ex2), hash(z * (x + y) + (x + y)))

    def test_get_variables(self):
        from . import make_vars, get_variables

        x, y, z = make_vars("x", "y", "z")
        self.assertEqual(get_variables(arg=x + y), ["x", "y"])
        self.assertEqual(get_variables(arg=[z - y, x + y]), ["x", "y", "z"])

    def test_rename_variables(self):
        from . import make_vars, rename_variables

        x, y, a, b = make_vars("x", "y", "a", "b")
        self.assertEqual(rename_variables(arg=x + y, d={"x": "b", "y": "a"}), b + a)
        self.assertEqual(
            rename_variables(arg=[x - y, x + y], d={"x": "b", "y": "a"}),
            [b - a, b + a],
        )

    def test_subs(self):
        from . import make_vars, subs

        x, y, a, b = make_vars("x", "y", "a", "b")
        self.assertEqual(str(subs(arg=x + y, smap={"x": b, "y": a})), "(b + a)")
        self.assertEqual(str(subs(arg=x + y, smap={x: b, y: a})), "(b + a)")
        self.assertEqual(
            str(subs(arg=[x + y, x - y], smap={"x": b, "y": a})[1]), "(b - a)"
        )
        self.assertEqual(str(subs(arg=[x + y, x - y], smap={x: b, y: a})[1]), "(b - a)")

    def test_relu_wrappers(self):
        from . import make_vars, leaky_relu, leaky_relup, relu, relup

        x, y = make_vars("x", "y")

        self.assertEqual(leaky_relu(0.0)(x), relu(x))
        self.assertEqual(leaky_relup(0.0)(x), relup(x))
        self.assertEqual(leaky_relu(0.1)(x + y), relu(x + y, 0.1))
        self.assertEqual(leaky_relup(0.1)(x + y), relup(x + y, 0.1))

        self.assertEqual(leaky_relu(0.0)(x * y), relu(x * y))
        self.assertEqual(leaky_relup(0.0)(x * y), relup(x * y))
        self.assertEqual(leaky_relu(0.1)(x * y + y), relu(x * y + y, 0.1))
        self.assertEqual(leaky_relup(0.1)(x * y + y), relup(x * y + y, 0.1))

    def test_dfun(self):
        from . import make_vars, dfun

        x, y = make_vars("x", "y")

        self.assertEqual(str(dfun("f", [x, y])), "(∂^0 f)")
        self.assertEqual(
            str(dfun(name="f", args=[x, y], didx=[(0, 3), (1, 5)])),
            "(∂^8 f)/(∂a0^3 ∂a1^5)",
        )

    def test_relational(self):
        from . import make_vars, lt, eq

        x, y = make_vars("x", "y")

        self.assertEqual(str(lt(x, y)), "(x < y)")
        self.assertEqual(str(eq(x, y)), "(x == y)")
        self.assertTrue("(x == 1.00" in str(eq(x, 1.0)))
        self.assertTrue("0000 == x)" in str(eq(1.0, x)))

        with self.assertRaises(TypeError) as cm:
            lt(1.0, 1.0)
        self.assertTrue(
            "At least one of the arguments of lt() must be an expression"
            in str(cm.exception)
        )

    def test_logical(self):
        from . import make_vars, logical_and, logical_or

        x, y = make_vars("x", "y")

        self.assertEqual(str(logical_and([x, y])), "logical_and(x, y)")
        self.assertEqual(str(logical_or([x, y])), "logical_or(x, y)")

    def test_select(self):
        from . import make_vars, select
        import numpy as np

        x, y = make_vars("x", "y")

        self.assertEqual(str(select(x, y, x)), "select(x, y, x)")

        self.assertTrue("select(1.0000" in str(select(1.0, y, x)))
        self.assertTrue("000, y, x)" in str(select(1.0, y, x)))

        self.assertTrue("select(1.0000" in str(select(1.0, y, 2.0)))
        self.assertTrue("000, y, 2.000" in str(select(1.0, y, 2.0)))

        with self.assertRaises(TypeError) as cm:
            select(1.0, 1.0, 1.0)
        self.assertTrue(
            "At least one of the arguments of select() must be an expression"
            in str(cm.exception)
        )

        with self.assertRaises(TypeError) as cm:
            select(1.0, y, np.longdouble(2.0))
        self.assertTrue(
            "The numerical arguments of select() must be all of the same type"
            in str(cm.exception)
        )

    def test_get_params(self):
        from . import make_vars, par, get_params

        x, y = make_vars("x", "y")

        self.assertEqual(get_params(x + y), [])
        self.assertEqual(get_params(x + par[42]), [par[42]])

        self.assertEqual(get_params([x + y, x - y]), [])
        self.assertEqual(get_params([x + par[42], par[1] - y]), [par[1], par[42]])

    def test_func_args(self):
        from . import make_vars, func_args

        x, y = make_vars("x", "y")

        fargs = func_args()
        self.assertEqual(fargs.args, [])
        self.assertFalse(fargs.is_shared)

        fargs = func_args([x, y])
        self.assertEqual(fargs.args, [x, y])
        self.assertFalse(fargs.is_shared)

        fargs = func_args(args=[x, y], shared=True)
        self.assertEqual(fargs.args, [x, y])
        self.assertTrue(fargs.is_shared)
